#pragma once

/**
 * FooIndecision:
 * @FOO_MAYBE:     Something maybe
 * @FOO_POSSIBLY:  Something possible
 *
 * The indecision type.
 **/

typedef enum {
    FOO_MAYBE,
    FOO_POSSIBLY,
} FooIndecision;

